

#include <cstdlib>
#include <cstdio>

#include <GL/glut.h>
#include "global.h"
#include "gravitycuda.h"


const int THREADSPERBLOCK = 64;
const int BLOCKS = N/THREADSPERBLOCK + (N % THREADSPERBLOCK == 0 ? 0 : 1);

__shared__ float4 shPosition[THREADSPERBLOCK];

__constant__ float KDAT = 0.405f;//405f;//4f;//8f;//8f;
__constant__ float BIGG = 0.001f;
__constant__ float EPS2 = 0.004;//25;//1f;


__device__ float3 bodybodyinteraction(float4 posi, float4 posj, float3 fi)
{
	float3 n;
	float DIST;
	float DIST3;

	n.x = posj.x - posi.x;
	n.y = posj.y - posi.y;

	DIST = sqrt(n.x * n.x + n.y * n.y + EPS2);

	DIST3 = DIST*DIST*DIST;

	fi.x += __fdividef(n.x * BIGG , DIST3);
	fi.y += __fdividef(n.y * BIGG , DIST3);

	return fi;
}


__device__ float3 tile_calculation(float4 pos, float3 forc)
{
	int i;

	#pragma unroll 64
	for (i = 0; i < blockDim.x; i++)//blockDim.x; i++) 
		forc = bodybodyinteraction(pos, shPosition[i], forc);
		
	return forc;
}



// using shared memory
__global__ void moveparticlesSharedOnDevice(FLOATVEC* posvbo, FLOATVEC *vel, float _DT, float3 _G, int _N, float3 _ATTRACT)
{
	
	int idx = blockIdx.x*blockDim.x + threadIdx.x;

	FLOATVEC posidx;
	FLOATVEC velidx;
	
	float3 ATTRFORCE;
	float3 forc;
	float DIST;

	int i,tile;

	if (idx<_N) 
	{
		posidx = posvbo[idx];

		// clear force
		//for(i=0; i < _N; i++)
	
		forc.x = 0;
		forc.y = 0;

		tile = 0;

		for (i = 0; i+threadIdx.x < _N; i += blockDim.x)				// i+threadIdx.x - poprawka do kodu nVidii
		{

			//=========== PARALLEL START HERE ============

			int pair_id = tile * blockDim.x + threadIdx.x;
			shPosition[threadIdx.x] = posvbo[pair_id];
			__syncthreads(); // fill out shPosition

			//=========== PARALLEL END HERE ==============


			//=========== PARALLEL START HERE ============

			forc = tile_calculation(posidx, forc);
			__syncthreads();	// calculate force acting on 

			//=========== PARALLEL END HERE ==============

			tile++;
		}


/*		for(int j=0; j < _N; j++)
		{
			forc = bodybodyinteraction(posidx,posvbo[j],forc);
		}
*/		
		DIST = sqrt(	(_ATTRACT.x-posidx.x)*(_ATTRACT.x-posidx.x) 
					+	(_ATTRACT.y-posidx.y)*(_ATTRACT.y-posidx.y));
		if(DIST)
		{
			ATTRFORCE.x = __fdividef(KDAT*((_ATTRACT.x-posidx.x)),DIST);
			ATTRFORCE.y = __fdividef(KDAT*((_ATTRACT.y-posidx.y)),DIST);
		}

		forc.x += _G.x + ATTRFORCE.x;
		forc.y += _G.y + ATTRFORCE.y;

		//forc = _G + ATTRFORCE;

		velidx = vel[idx];

		velidx.x = velidx.x + forc.x * _DT;
		velidx.y = velidx.y + forc.y * _DT;

		posidx.x += velidx.x * _DT;
		posidx.y += velidx.y * _DT;

//		if(posidx.x > 1) { posidx.x = 1.0f; velidx.x = -velidx.x; }
//		if(posidx.y > 1) { posidx.y = 1.0f; velidx.y = -velidx.y; }
//		if(posidx.x < -1) { posidx.x = -1.0f; velidx.x = -velidx.x; }
//		if(posidx.y < -1) { posidx.y = -1.0f; velidx.y = -velidx.y; }

		posvbo[idx] = posidx;				// !!!!!!!!!!!!!!
		vel[idx] = velidx;
	}

}


__global__ void moveparticlesOnDevice(FLOATVEC* posvbo, FLOATVEC *vel, float _DT, float3 _G, int _N, float3 _ATTRACT)
{
	int idx = blockIdx.x*blockDim.x + threadIdx.x;

	FLOATVEC posidx;
	FLOATVEC velidx;
	
	float3 ATTRFORCE;
	float3 forc;
	float3 n;
	float DIST;

	if (idx<_N) 
	{
		posidx = posvbo[idx];

		// clear force
		//for(i=0; i < _N; i++)
	
		forc.x = 0;
		forc.y = 0;

		for(int j=0; j < _N; j++)
		{
			DIST = sqrt(	(posidx.x - posvbo[j].x)*(posidx.x - posvbo[j].x) 
						+	(posidx.y - posvbo[j].y)*(posidx.y - posvbo[j].y) + EPS2);

			if(DIST)
			{
				n.x = __fdividef((posidx.x - posvbo[j].x) , DIST);
				n.y = __fdividef((posidx.y - posvbo[j].y) , DIST);
	
				// fg = G*m1m2/r**2

				forc.x -= __fdividef(n.x * BIGG , DIST*DIST);
				forc.y -= __fdividef(n.y * BIGG , DIST*DIST);
			
			}
		}
		
		DIST = sqrt(	(_ATTRACT.x-posidx.x)*(_ATTRACT.x-posidx.x) 
					+	(_ATTRACT.y-posidx.y)*(_ATTRACT.y-posidx.y));
		if(DIST)
		{
			ATTRFORCE.x = __fdividef(KDAT*((_ATTRACT.x-posidx.x)),DIST);
			ATTRFORCE.y = __fdividef(KDAT*((_ATTRACT.y-posidx.y)),DIST);
		}

		forc.x += _G.x + ATTRFORCE.x;
		forc.y += _G.y + ATTRFORCE.y;

		velidx = vel[idx];

		velidx.x = velidx.x + forc.x * _DT;
		velidx.y = velidx.y + forc.y * _DT;

		posidx.x += velidx.x * _DT;
		posidx.y += velidx.y * _DT;

//		if(posidx.x > 1) { posidx.x = 1.0f; velidx.x = -velidx.x; }
//		if(posidx.y > 1) { posidx.y = 1.0f; velidx.y = -velidx.y; }
//		if(posidx.x < -1) { posidx.x = -1.0f; velidx.x = -velidx.x; }
//		if(posidx.y < -1) { posidx.y = -1.0f; velidx.y = -velidx.y; }

		posvbo[idx] = posidx;				// !!!!!!!!!!!!!!
		vel[idx] = velidx;
	}


}



FLOATVEC *vel_device;

FLOATVEC *pos_host;
FLOATVEC *vel_host;
FLOATVEC *forc_host;

float *mas_host;

void allocate_memory_cuda(int _N)
{
	printf("alokacja... \n");

	size_t size = _N*sizeof(float4);
	size_t sizefloat = _N*sizeof(float);

	pos_host = (FLOATVEC *) malloc(size);
	vel_host = (FLOATVEC *) malloc(size);
	forc_host = (FLOATVEC *) malloc(size);
	mas_host = (float *) malloc(sizefloat);

	cudaMalloc((void **) &vel_device, size);
}

void deallocate_memory_cuda(void)
{
	printf("dealokacja... \n");

	free(pos_host);
	free(vel_host);
	free(forc_host);
	free(mas_host);
	
	cudaFree(vel_device); 
}

void copy_memory_cuda(int _N)
{
	cudaMemcpy(vel_device, vel_host, sizeof(float4)*_N, cudaMemcpyHostToDevice);
}


void moveParticlesCUDA(FLOATVEC* posvbo, float _DT, float3 _G, int _N, float3 _ATTRACT)
{


//	printf("blocks: %d,", blocks);
	moveparticlesSharedOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, vel_device, _DT, _G, _N, _ATTRACT);
	//moveparticlesOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, vel_device, _DT, _G, _N, _ATTRACT);
	cudaThreadSynchronize();

}