
#include <GL/glew.h>
#include <GL/glut.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <vector>
#include <cmath>
using namespace std;

#include "ballscpu.h"
#include "ballscuda.h"
#include "dumpgl.h"
#include "global.h"

// cuda includes

#include <cuda_runtime.h>
//#include <cutil_inline.h>
//#include <cutil_gl_error.h>
#include <cuda_gl_interop.h>
#include <vector_types.h>

// 

const int WIDTH = 600;
const int HEIGHT = 600;
const float DT = 0.010f;
float3 G;
FLOATVEC ATTRACT;
int selected_ball_i = -1;
void change_ball_position(void);
float4 selected_ball_pos;
float4 ball_0_pos;


// vbo variables
GLuint vbo;
GLuint vbocolors;

// switches
int doclear = 1;

void createVBO(GLuint* _vbo, int _mode)
{
    // create buffer object
    glGenBuffers(1, _vbo);
    glBindBuffer(GL_ARRAY_BUFFER, *_vbo);

    // initialize buffer object
    unsigned int size = N * sizeof(FLOATVEC);						// pos
    glBufferData(GL_ARRAY_BUFFER, size, 0, _mode);//GL_DYNAMIC_DRAW);

    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // register buffer object with CUDA
    cudaGLRegisterBufferObject(*_vbo);


}


void deleteVBO(GLuint* _vbo)
{
    glBindBuffer(1, *_vbo);
    glDeleteBuffers(1, _vbo);
    cudaGLUnregisterBufferObject(*_vbo);
	
	*_vbo = 0;
}


void initcolorbuffer(void)
{

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbocolors);
	FLOATVEC *ptr = (FLOATVEC*)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);
	if(ptr)
	{
		const float maxc = 0.3f;
		const float redf = 4.0f;
		const float greenf = 3.0f;
		const float bluef = 0.0f;

		float r;
		for(int k=0; k<N; k++)
		{
			
			r = maxc*rand()/float(RAND_MAX);
			ptr[k].x = redf*r;
			ptr[k].y = greenf*r;
			ptr[k].z = bluef*r;
			ptr[k].w = 0.85f;
		}
	}
	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);
}

void init(void)
{	
    // initialize necessary OpenGL extensions
    glewInit();
    if (! glewIsSupported("GL_VERSION_2_0 ")) {
        fprintf(stderr, "ERROR: Support for necessary OpenGL extensions missing.");
        fflush(stderr);
		exit(0);
        //return CUTFalse;
    }

    // create VBO
    createVBO(&vbo, GL_DYNAMIC_DRAW);//STREAM_DRAW);
	
	// colors vbo + init
	createVBO(&vbocolors, GL_STATIC_DRAW);
	
	initcolorbuffer();


	allocate_memory_cuda(N);
	initparticles();
	copy_memory_cuda(N);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo);
	FLOATVEC *ptrp = (FLOATVEC*)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);
	if(ptrp)
	{
		for(int k=0; k<N; k++)
		{
			ptrp[k].x = pos_host[k].x;
			ptrp[k].y = pos_host[k].y;
			ptrp[k].z = pos_host[k].z;
			//ptrp[k].w = 0;
		}
	}
	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);

	G.x = 0;
	G.y = -2.5;
	G.z = 0;
}

/* GLUT callback Handlers */

static void resize(int width, int height)
{
    const float ar = (float) width / (float) height;
    
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
#define w 1.0
	gluOrtho2D(-w,w,-w,w);//(NEW)  Define our viewing area
 
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity() ;
}

static void display(void)
{
    const double t = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
    const double a = t*90.0;

	if(doclear)
	{
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}    
		
	glColor3d(1,0,0);

	const float pointsizef = 16.0f;
	const float redf = 4.0f;
	const float greenf = 3.0f;


	glPointSize(1.8);

  // render from the vbo



//#define VIS_GPU
//#define VIS_CPU


//glDepthFunc(GL_LEQUAL);
//glAlphaFunc(GL_GREATER, 0.6);
//glEnable(GL_ALPHA_TEST);


#ifdef DO_GPU

#define _0_POINT_VIS
//#define SEL_POINT_VIS

//	glEnable(GL_POINT_SPRITE);
//	glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE); 

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(12.8);

	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexPointer(3, GL_FLOAT, sizeof(float4),0);
	glBindBuffer(GL_ARRAY_BUFFER, vbocolors);
	glColorPointer(4, GL_FLOAT, 0, 0);
	

    glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glDrawArrays(GL_POINTS, 1, N-1);
    glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);


//	glDisable(GL_POINT_SPRITE);
#endif


#ifdef DO_CPU
	glPointSize(2.6);
	glColor4f(0.4,0.4,0.4,1.0);
	

	for(int i=0; i<N; i++)
	{
		glPushMatrix();
		glTranslatef(pos_host[i].x, pos_host[i].y, pos_host[i].z);
		glColor3f(0.3,0.3,0.3);
		//glutWireSphere(R_host[i],7,5);
		//glutWireSphere(R_host[i],7,5);
		glutWireSphere(0.98*R_host[i],9,4);

		if (i == selected_ball_i)
		{
			glColor4f(0.3,0.3,0.3,0.5);
			glutSolidSphere(0.98*R_host[i],9,4);
		}

		//glPointSize(2.6);
		//glVertex2f( pos_host[i].x, pos_host[i].y );
		glPopMatrix();
	}
	

#endif

	#ifdef SEL_POINT_VIS
	//selected_ball_poz
	glPointSize(55.0);
	glColor4f(0.0,0,1.0,1.0);
	#define e 0.06
	glBegin(GL_LINE_STRIP);
		glVertex2f( selected_ball_pos.x+e, selected_ball_pos.y+e );
		glVertex2f( selected_ball_pos.x+e, selected_ball_pos.y-e );
		glVertex2f( selected_ball_pos.x-e, selected_ball_pos.y-e );
		glVertex2f( selected_ball_pos.x-e, selected_ball_pos.y+e );
		glVertex2f( selected_ball_pos.x+e, selected_ball_pos.y+e );
	glEnd();
	#endif


	// we draw it because it should be bigger on the screen
	#ifdef _0_POINT_VIS
	//selected_ball_poz
	glPointSize(150.0);
	//glColor4f(1.0,0.9,0.8,1.0);
	//glBegin(GL_POINTS);
	//glVertex2f( ball_0_pos.x, ball_0_pos.y );
	//glEnd();
	glColor4f(0.2,0.18,0.0,1.0);
	glPushMatrix();
		glTranslatef( ball_0_pos.x, ball_0_pos.y, 0 );
		glutSolidSphere( 0.17,10,10);
	glPopMatrix();
	#endif

	glutSwapBuffers();
	
}

float wasg;
static void key(unsigned char key, int x, int y)
{
    switch (key) 
    {
        case 27 : 
        case 'q':
			deleteVBO(&vbo);
			deleteVBO(&vbocolors);
			deallocate_memory_cuda();
            exit(0);
            break;
        case ' ':
			doclear = 1-doclear;
			break;

		case 'g':
			if(G.y!=0)
			{
				wasg = G.y;
				G.y = 0;
			}
			else
			{
				G.y = wasg;
			}
			break;

	}

    glutPostRedisplay();
}


static void idle(void)
{

	static int step = 0;

	#define NUMAVERAGE 5

	static vector<float> averagetcpu(NUMAVERAGE);
	static vector<float> averagetgpu(NUMAVERAGE);



//	sg_glDumpWindow(WIDTH,HEIGHT);


//	DWORD t1c,t2c,t1g,t2g;

	
	//t1c=GetTickCount();
//	printf("executing CPU...");

   #ifdef DO_CPU
   moveparticles(DT,G,ATTRACT);
   #endif
	//t2c=GetTickCount();
//	printf(".. end, time(tick) = %f \n",	(t2c-t1c)*0.001f);
//	t1g=GetTickCount();
	//printf("GPU...");
		// map OpenGL buffer object for writing from CUDA
		// 
   #ifdef DO_GPU
    
	
		FLOATVEC *dptr;
		cudaGLMapBufferObject((void**)&dptr, vbo);
		moveParticlesCUDA(dptr,DT,G,N);//selected_ball_gpu_i);
		readposCUDA(dptr,N,&selected_ball_pos, &ball_0_pos);
	// unmap buffer object
	    cudaGLUnmapBufferObject(vbo);

   #endif

	
//	t2g=GetTickCount();
	//printf(".. end, time(tick) = %f \n",	(t2g-t1g)*0.001f);


/*
	if(step < NUMAVERAGE)		// 0..9
	{
		averagetcpu[step] = (t2c-t1c) * 0.001f;
		averagetgpu[step] = (t2g-t1g) * 0.001f;
		step++;
	}
	else
	{
		for(int i=1; i<NUMAVERAGE; i++)
		{
			averagetcpu[0] += averagetcpu[i];
			averagetgpu[0] += averagetgpu[i];
		}

		cout << averagetcpu[0] / float(NUMAVERAGE) << "\t" << averagetgpu[0] / float(NUMAVERAGE) << endl;
		//cout << averagetcpu[0] / averagetgpu[0] << endl;
		step=0;
	}
	

//	static int s=0;
//	s++;
//	if(!(s%1000))
    //sg_glDumpWindow(WIDTH,HEIGHT);

*/


	glutPostRedisplay();
}


	/**
	 * Mouse
	 *
	 */


void change_ball_selection(void)
{
	float dist;
	for(int i=0; i<N; i++)
	{
		//glTranslatef(pos_host[i].x, pos_host[i].y, pos_host[i].z);

		dist = sqrt (	(ATTRACT.x - pos_host[i].x)*(ATTRACT.x - pos_host[i].x) +
						(ATTRACT.y - pos_host[i].y)*(ATTRACT.y - pos_host[i].y) +
						(ATTRACT.z - pos_host[i].z)*(ATTRACT.z - pos_host[i].z));
		if(dist - R_host[i] < 0)
		{
			selected_ball_i = i;
			break;
		}
	}

	// CUDA

	FLOATVEC *dptr;
	cudaGLMapBufferObject((void**)&dptr, vbo);
	findselectCUDA(dptr,ATTRACT,N);
    cudaGLUnmapBufferObject(vbo);
}

void clear_ball_selection(void)
{
	selected_ball_i = -1;

	// gpu

}

void change_ball_position(void)
{
	pos_host[selected_ball_i].x = ATTRACT.x;
	pos_host[selected_ball_i].y = ATTRACT.y;


	// CUDA
	FLOATVEC *dptr;
	cudaGLMapBufferObject((void**)&dptr, vbo);
	changePosCUDA(dptr,ATTRACT);
	cudaGLUnmapBufferObject(vbo);

}

float posx[2]={0}, posy[2]={0}, translatex[2]={0}, translatey[2]={0};
int mousedown[2] = {0}, prevx[2]={0}, prevy[2]={0};

#define L 0
#define R 1

void Mouse (int button, int state, int x, int y)
{
	int K;

	if (button == GLUT_LEFT_BUTTON) K=L;
	else if (button == GLUT_RIGHT_BUTTON) K=R;

	if (state == GLUT_DOWN)
	{
		mousedown[K] = 1;

		ATTRACT.x = 2.0 * (x/float(WIDTH) - 0.5);
		ATTRACT.y = -2.0 * (y/float(HEIGHT) - 0.5);

		// loop over all balls
		// find one which lays at the point 
		// mouse was clicked at

		change_ball_selection();

	} else if (state == GLUT_UP)
	{
		mousedown[K] = 0;
	//	clear_ball_selection();
	}

}

/* Mouse Motion */
void Motion (int x, int y)
{
	if (mousedown[L])
	{

		ATTRACT.x = 2.0 * (x/float(WIDTH) - 0.5);
		ATTRACT.y = -2.0 * (y/float(HEIGHT) - 0.5);

		// loop over all balls
		// find one which lays at the point 
		// mouse was clicked at

		change_ball_position();
		glutPostRedisplay();
	}

	if (mousedown[R])
	{
	}
}



int main(int argc, char *argv[])
{
	cudaGLSetGLDevice(0);

    glutInit(&argc, argv);
	
	glutInitWindowSize(WIDTH,HEIGHT);
    glutInitWindowPosition(100,100);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);// | GLUT_DEPTH);

    glutCreateWindow("Granular matter, CUDA");

	init();


    glutReshapeFunc(resize);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);

	glutIdleFunc(idle);

    glClearColor(1,1,1,1);

    glutMainLoop();

    return EXIT_SUCCESS;
}

